﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;

using As.UCCNCplugins.Forms;

using static UC100;

namespace As.UCCNCplugins
{
    /// <summary>
    /// UCCNCplugin - UCCNC plugin.
    /// </summary>
    /// <remarks>Class name must be UCCNCplugin to work, do not change it.</remarks>
    public class UCCNCplugin : IDisposable
    {
        #region Assembly data
        /// <summary>
        /// AUTHOR - My name here.
        /// </summary>
        public const string AUTHOR = "A.H.M. Steenveld";

        /// <summary>
        /// UCCNCplugin - static .ctor, inisialising data available to each instance.
        /// </summary>
        /// <remarks>This is activated once (by the OS loader) just before the first instance is created.</remarks>
        static UCCNCplugin()
        {
            InitialiseAssemblyData(ref Home, ref Name, ref Version, ref Product);
            InitialiseLogger();
        }

        /// <summary>
        /// InitialiseAssemblyData - get some well known assembly data more easely available.
        /// </summary>
        /// <param name="home">reference to the home path variable</param>
        /// <param name="name">reference to the name variable</param>
        /// <param name="version">reference to the version variable</param>
        /// <param name="product">reference to the version variable</param>
        static void InitialiseAssemblyData(ref string home, ref string name, ref Version version, ref string product)
        {
            try
            {
                // see https://stackoverflow.com/questions/52797/how-do-i-get-the-path-of-the-assembly-the-code-is-in/283917#283917
                var codeBase = Assembly.GetExecutingAssembly().CodeBase;
                var uri = new UriBuilder(codeBase);
                var path = Uri.UnescapeDataString(uri.Path);

                home = Path.GetDirectoryName(path);
                name = Path.GetFileNameWithoutExtension(path);

                version = typeof(UCCNCplugin).Assembly.GetName().Version;

                product =
                    ((AssemblyProductAttribute)typeof(UCCNCplugin)
                    .Assembly
                    .GetCustomAttributes(typeof(AssemblyProductAttribute), true)[0])
                    .Product;
            }
            catch (Exception x)
            {
                LogWriteLine($"ERROR: Unhandled exception: {x}");
            }
        }

        /// <summary>
        /// Home - Home path of this assembly
        /// </summary>
        internal static readonly string Home;

        /// <summary>
        /// Name - Name of this assembly
        /// </summary>
        public static readonly string Name;

        /// <summary>
        /// Version - Version of this assembly
        /// </summary>
        public static readonly Version Version;

        /// <summary>
        /// Product - Short description of this assembly
        /// </summary>
        public static readonly string Product;
        #endregion Assembly data

        #region Instance handling
        /// <summary>
        /// UCCNCplugin - .ctor, initialise a new instance.
        /// </summary>
        /// <remarks>
        /// Just a word of precausion.
        /// 
        /// This plugin is for the development environment only, testing UCCNC API's etc.
        /// It is (inherently) *not safe* to load it by default in a productions environment.
        /// 
        /// One of the risks is the following. By default the log is using a queue.
        /// When the application is not changing default log configuration that queue can
        /// overflow and cause memory problems.
        /// 
        /// So, a great environment to find out how things work, just what is needed while developing.
        /// But for release and productions all the developing and debug facilities can be a risk.
        /// </remarks>
        public UCCNCplugin()
        {
            try
            {
                InitializeIDisposable();

                LogWriteLine($"Info: UCCNC plugin: action = \"Create\"; name = {Name}; version = {Version}; time = {DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss")}");
                clock = new System.Threading.Timer(OnTick);

                InitialiseApp();

                LogWriteLine("Warn: UCCNC plugin: This plugin is for the development environment only. It is (inherently) *not safe* to load it by default in a productions environment.");
            }
            catch (Exception x)
            {
                LogWriteLine($"FATAL: UCCNC plugin: problem = \"Unhandled application exception, bailing out\"; message = \"{x.Message}\" ");
            }
        }

        /// <summary>
        /// ~UCCNCplugin - .dtor, finalise an instance.
        /// </summary>
        /// <remarks>
        /// This is activated once per instance (by the OS loader) just before the garbadge collector destroys the instance.
        /// Do not use in your code!
        /// </remarks>
        /// <see cref="Finalise"/>
        ~UCCNCplugin() { Dispose(false); }

        /// <summary>
        /// Finalise - terminate all activities, release all resorces.
        /// </summary>
        /// <remarks>Entry point at instance level for cleaning up. Use this in your code.</remarks>
        void Finalise()
        {
            LogWriteLine("Debug: UCCNC plugin: action = \"Finalise\"");
            if (clock != null)
            {
                clock.Dispose();
                clock = null;
            }
            FinaliseApp();
            FinaliseLogger();
        }

        /// <summary>
        /// UC - The Plugin interface object which contains methods, properties and more to get things done.
        /// </summary>
        public Plugininterface.Entry UC { get; private set; } = null;
        #endregion Instance handling

        #region Clock
        /// <summary>
        /// CLOCK_TIME_DUE - duration of the first waiting cycle.
        /// </summary>
        const int CLOCK_TIME_DUE = 400; // msec

        /// <summary>
        /// CLOCK_TIME - duration of subsequent waiting cycles.
        /// </summary>
        const int CLOCK_TIME = 200; // msec

        /// <summary>
        /// clock - for notices at regular intervals.
        /// </summary>
        System.Threading.Timer clock;

        public delegate void POnClockTick();

        /// <summary>
        /// OnClockTick - list of callbacks listening to clock events.
        /// </summary>
        /// <remarks>The first callback entering will start the clock, the last callback leaving will stop the clock</remarks>
        public event POnClockTick OnClockTick
        {
            add
            {
                if (on_clock_tick == null) clock?.Change(CLOCK_TIME_DUE, CLOCK_TIME);
                on_clock_tick += value;
            }
            remove
            {
                on_clock_tick -= value;
                if (on_clock_tick == null) clock?.Change(Timeout.Infinite, Timeout.Infinite);
            }
        }

        /// <summary>
        /// on_clock_tick - container for OnClockTick
        /// </summary>
        event POnClockTick on_clock_tick;

        /// <summary>
        /// OnTick - handler for clock events.
        /// </summary>
        /// <param name="state"></param>
        void OnTick(object state) { on_clock_tick?.Invoke(); }
        #endregion Clock

        #region Logger
        /// <summary>
        /// InitialiseLogger - Check if logging to file is available.
        /// </summary>
        static void InitialiseLogger()
        {
            var path = Path.Combine(Home, $"{Name}.log");
            if (File.Exists(path))
            {
                log_path = path;
                Exception e = LogOpen();
                if (e != null) LogWriteLine($"UCCNC plugin: open log = failed; {e.Message}; path = \"{path}\"");
                else if (log == null) LogWriteLine($"UCCNC plugin: open log = failed; path = \"{path}\"");
            }
            else
            {
                LogWriteLine($"UCCNC plugin: write log = disabled; path = \"{path}\"");
            }
        }

        /// <summary>
        /// FinaliseLogger - Close the log file.
        /// </summary>
        static void FinaliseLogger()
        {
            LogClose();
        }

        /// <summary>
        /// LogUseQueue - flag to indicate the use of a queue of messages for further processing.
        /// </summary>
        static bool LogUseQueue
        {
            get { lock (key_logger) { return _log_use_queue; } }
            set
            {
                lock (key_logger)
                {
                    _log_use_queue = value;
                    if (value) LogClear();
                }
            }
        }

        /// <summary>
        /// _log_use_queue - container for LogUseQueue
        /// </summary>
        static bool _log_use_queue = true;

        /// <summary>
        /// LogVerbose - True: log many details. False: log problem details.
        /// </summary>
#if DEBUG
        static public bool LogVerbose { get; set; } = true;
#else
        static public bool LogVerbose { get; set; } = false;
#endif

        /// <summary>
        /// key_logger - Lock used by threads for access to logger resources.
        /// </summary>
        static object key_logger = new object();

        /// <summary>
        /// log_messages - queue of log messages waiting to be displayed (or disposed) by the application.
        /// </summary>
        static Queue<string> log_messages = new Queue<string>();

        /// <summary>
        /// log_open_tried - True: Opening the log file the first time was done.
        /// </summary>
        static bool log_open_tried = false;

        /// <summary>
        /// log_path - path to the log file
        /// </summary>
        /// <remarks>
        /// When log_open_tried is True:
        /// If log_path is null then there was no log file or there were problems opening it.
        /// If log_path is not null it is possible to open the log file without problems.
        /// </remarks>
        static string log_path = null;

        /// <summary>
        /// log - writer to a log file.
        /// </summary>
        /// <remarks>
        /// When log_path is not null the log will (re)opened when needed.
        /// </remarks>
        static TextWriter log = null;

        /// <summary>
        /// LogOpen - Open log file, if failing set a flag not to try it again.
        /// </summary>
        /// <returns>Event when there are problems during opening, null otherwise</returns>
        /// <remarks>Cannot use LogWriteLine</remarks>
        static Exception LogOpen()
        {
            Exception e = null;
            if ((log != null) || (log_path == null)) return e;
            lock (key_logger)
            {
                try
                {
                    log = File.AppendText(log_path);
                    if (!log_open_tried)
                    {
                        foreach (var m in log_messages) log.WriteLine(m);
                    }
                }
                catch (Exception x)
                {
                    log_path = null;
                    e = x;
                    if (log != null)
                    {
                        try { log.Dispose(); } catch { /* never mind */ }
                        log = null;
                    }
                }
                finally
                {
                    log_open_tried = true;
                }
            }
            return e;
        }

        /// <summary>
        /// LogCLose - Close the log file resource.
        /// </summary>
        static void LogClose()
        {
            lock (key_logger)
            {
                if (log != null)
                {
                    log.Close();
                    log.Dispose();
                    log = null;
                }
            }
        }

        /// <summary>
        /// LogWriteLine - Write a line to the log, reopen log file if needed (and possible).
        /// </summary>
        /// <param name="message">log message</param>
        public static void LogWriteLine(string message)
        {
            if (message == null) return;
            lock (key_logger)
            {
                if (LogUseQueue) log_messages.Enqueue(message);
                if (log == null) LogOpen();
                if (log != null)
                {
                    log.WriteLine(message);
                    log.Flush();
                }
            }
        }

        /// <summary>
        /// LogTryRead - Get the next log message for furthere processing (e.g.: display it on screen).
        /// </summary>
        /// <param name="message">The next message from the log queue, if available</param>
        /// <returns>True if a message was available, False otherwise</returns>
        public static bool LogTryRead(ref string message)
        {
            var result = false;
            lock (key_logger)
            {
                if (0 < log_messages.Count)
                {
                    result = true;
                    message = log_messages.Dequeue();
                }
            }
            return result;
        }

        /// <summary>
        /// LogClear - Remove all log messages.
        /// </summary>
        public static void LogClear()
        {
            lock (key_logger) log_messages.Clear();
        }
        #endregion Logger

        #region IDisposable
        void InitializeIDisposable()
        {
            disposed = false;
        }

        // sometimes required for other interfaces.
        public event EventHandler Disposed;

        bool disposed;

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            LogWriteLine($"Debug: UCCNC plugin: action = Dispose({disposing})");
            if (!disposed)
            {
                if (disposing)
                {
                    Finalise();
                    Disposed?.Invoke(this, new EventArgs());
                }
                disposed = true;
            }
        }
        #endregion

        #region IUCCNCplugin events
        /// <summary>
        /// Init_event - Called when the plugin is initialised.
        /// </summary>
        /// <param name="UC">The Plugin interface object which contains all functions prototypes for calls and callbacks</param>
        public void Init_event(Plugininterface.Entry UC)
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Init_event(UC)");
            if (UC != null) LogWriteLine("Warn: UCCNC plugin: Init_event; action = \"init repeated event\"");
            try
            {
                this.UC = null;
                firstrun = true;
                this.UC = UC;
                app_init();
            }
            catch (Exception x)
            {
                LogWriteLine($"ERROR: UCCNC plugin: Init_event; problem = \"not handled here\"; message = \"{x.Message}\"");
                throw;
            }
        }

        /// <summary>
        /// Shutdown_event - Called when the UCCNC software is closing
        /// </summary>
        public void Shutdown_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Shutdown_event()");
            if (UC == null) LogWriteLine("Warn: Shutdown_event: action = skip; reason = \"UC = null, already in shutdown state\"");
            app_Shutdown();
        }

        /// <summary>
        /// Properties - Called when the plugin is loaded.
        /// </summary>
        /// <param name="value">set the details of the plugin in this object</param>
        /// <returns>The Properties object</returns>
        public Plugininterface.Entry.Pluginproperties Getproperties_event(Plugininterface.Entry.Pluginproperties value)
        {
            value.author = AUTHOR;
            value.pluginname = Name;
            value.pluginversion = Version.ToString();
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Getproperties_event(author = \"{value.author}\", pluginname = \"{value.pluginname}\", pluginversion = {value.pluginversion})");
            return value;
        }

        /// <summary>
        /// Configure_event - Called from UCCNC when the user presses the Configuration button in the Plugin configuration menu
        /// </summary>
        /// <remarks>Typically the plugin configuration window is shown to the user</remarks>
        public void Configure_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Configure_event()");
            app_Configure();
        }

        /// <summary>
        /// Startup_event - Called from UCCNC when the plugin is loaded and started
        /// </summary>
        public void Startup_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Startup_event()");
            app_Startup();
        }

        /// <summary>
        /// Showup_event - Called when the Pluginshowup(string Pluginfilename); function is executed in the UCCNC
        /// </summary>
        public void Showup_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Showup_event()");
            app_Showup();
        }

        /// <summary>
        /// Informplugin_event - This is a direct function call addressed to this plugin dll
        /// </summary>
        /// <param name="value">Object with message details</param>
        /// <returns>returning message details</returns>
        /// <remarks>The function can be called by macros or by another plugin</remarks>
        public object Informplugin_event(object value)
        {
            var msg = (value is string) ? $"\"{value}\"" : $"{value}";
            object result = null; // app_Informplugin(value);
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Informplugin_event(Message = \"{msg}\") = \"{result}\"");
            return result;
        }

        /// <summary>
        /// Informplugins_event - This is a function call made to all plugin dll files
        /// </summary>
        /// <param name="value">Message details</param>
        /// <remarks>The function can be called by macros or by another plugin</remarks>
        public void Informplugins_event(object value)
        {
            var msg = (value is string) ? $"\"{value}\"" : $"{value}";
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Informplugins_event(Message = \"{msg}\")");
            // app_Informplugins(value);
        }

        /// <summary>
        /// Buttonpress_event - Called when the user presses a button on the UCCNC GUI or if a Callbutton function is executed
        /// </summary>
        /// <param name="buttonnumber">ID of the caller button</param>
        /// <param name="onscreen">true if the button was pressed on the GUI and is false if the Callbutton function was called</param>
        public void Buttonpress_event(int buttonnumber, bool onscreen)
        {
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Buttonpress_event(buttonnumber = {buttonnumber}, onscreen = {onscreen})");
            // app_Buttonpress(buttonnumber, onscreen);
        }

        /// <summary>
        /// Toolpathclick_event - Called when the user clicks the toolpath viewer
        /// </summary>
        /// <param name="X">click X coordinate in the model space of the toolpath viewer</param>
        /// <param name="Y">click Y coordinate in the model space of the toolpath viewer</param>
        /// <param name="Istopview">true if the toolpath viewer is rotated into the top view, the passed coordinates are only precise when the top view is selected</param>
        public void Toolpathclick_event(double X, double Y, bool Istopview)
        {
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Toolpathclick_event(X = {X:0.000}, Y = {Y:0.000}, Istopview = {Istopview})");
            // app_Toolpathclick(X, Y, Istopview);
        }

        /// <summary>
        /// Textfieldclick_event - Called when the user clicks and enters a Textfield on the screen
        /// </summary>
        /// <param name="labelnumber">ID of the accessed Textfield</param>
        /// <param name="Ismainscreen">true is the Textfield is on the main screen and false if it is on the jog screen</param>
        public void Textfieldclick_event(int labelnumber, bool Ismainscreen)
        {
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Textfieldclick_event(labelnumber = {labelnumber}, Ismainscreen = {Ismainscreen})");
            // app_Textfieldclick(labelnumber, Ismainscreen);
        }

        /// <summary>
        /// Textfieldtexttyped_event - Called when the user enters text into the Textfield and it gets validated
        /// </summary>
        /// <param name="labelnumber">ID of the accessed Textfield</param>
        /// <param name="Ismainscreen">true is the Textfield is on the main screen and false if it is on the jog screen</param>
        /// <param name="text">text entered and validated by the user</param>
        public void Textfieldtexttyped_event(int labelnumber, bool Ismainscreen, string text)
        {
            if (LogVerbose) LogWriteLine($"Debug: UCCNC plugin: Textfieldclick_event(labelnumber = {labelnumber}, Ismainscreen = {Ismainscreen}, text = \"{text}\")");
            // app_Textfieldtexttyped(labelnumber, Ismainscreen, text);
        }

        /// <summary>
        /// Cyclethreadstart_event - Called when the user presses the Cycle start button and before the Cycle starts
        /// </summary>
        /// <remarks>
        /// This event may be used to show messages or do actions on Cycle start
        /// For example to cancel the Cycle if a condition met before the Cycle starts with calling the Button code 130 Cycle stop
        /// </remarks>
        public void Cyclethreadstart_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Cyclethreadstart_event()");
            // app_Cyclethreadstart();
        }

        /// <summary>
        /// Cyclethreadstop_event - Called when the user presses the Cycle stop button and after the Cycle stops
        /// </summary>
        public void Cyclethreadstop_event()
        {
            if (LogVerbose) LogWriteLine("Debug: UCCNC plugin: Cyclethreadstop_event()");
            // app_Cyclethreadstop();
        }
        #endregion IUCCNCplugin events

        #region IUCCNCplugin loop events
        /// <summary>
        /// key - object for synchronising threads.
        /// </summary>
        public readonly object Key = new object();

        /// <summary>
        /// loopfirstrun - flag to indicate is the first fun of the loop has to happen yet.
        /// </summary>
        public bool firstrun
        {
            get { lock (Key) { return _firstrun; } }
            set { lock (Key) { _firstrun = value; } }
        }
        bool _firstrun = true;

        /// <summary>
        /// loopstop - flag to indicate that the loop should stop
        /// </summary>
        public bool loopstop
        {
            get { lock (Key) { return _loopstop; } }
            set { lock (Key) { _loopstop = value; } }
        }
        bool _loopstop = false;

        /// <summary>
        /// loopactive - flag to indicate that the loop is currently taking a cycle.
        /// </summary>
        public bool loopworking
        {
            get { lock (Key) { return _loopworking; } }
            set { lock (Key) { _loopworking = value; } }
        }
        bool _loopworking = true;

        /// <summary>
        /// Loop_event - Called in a loop with a 25Hz interval
        /// </summary>
        public void Loop_event()
        {
            if (loopstop) return;
            try
            {
                loopworking = true;
                if (firstrun) firstrun = app_LoopFirstCycle();
                app_LoopCycle();
            }
            catch (Exception x) { app_LoopException(x); }
            finally { loopworking = false; }
        }
        #endregion IUCCNCplugin loop events

        #region Position
        /// <summary>
        /// IsMoving - True: movement in progress on at least one axis. False: no movement on any of the axis.
        /// </summary>
        public bool IsMoving => UC?.IsMoving() ?? false;

        /// <summary>
        /// X - Current X position as text
        /// </summary>
        public string X => UC?.Getfield(true, 226) ?? "-";

        /// <summary>
        /// Y - Current Y position as text
        /// </summary>
        public string Y => UC?.Getfield(true, 227) ?? "-";

        /// <summary>
        /// Z - Current Z position as text
        /// </summary>
        public string Z => UC?.Getfield(true, 228) ?? "-";

        /// <summary>
        /// x - Current x position
        /// </summary>
        public double x => UC?.Getfielddouble(true, 226) ?? double.NaN;

        /// <summary>
        /// y - Current y position
        /// </summary>
        public double y => UC?.Getfielddouble(true, 227) ?? double.NaN;

        /// <summary>
        /// z - Current z position
        /// </summary>
        public double z => UC?.Getfielddouble(true, 228) ?? double.NaN;
        #endregion Position

        #region Application
        void InitialiseApp()
        {
            // switch verbose logging on to see what is happening.
            LogVerbose = true;
        }

        void FinaliseApp()
        {
            CloseForm();
        }
        #endregion Application

        #region Forms related events
        Form main = null;

        void app_Startup()
        {
            OpenForm();
        }

        void app_Showup()
        {
            OpenForm();
            main?.BringToFront();
        }

        void app_Shutdown()
        {
            Finalise(); // Release *all* resources.
        }

        void app_Configure()
        {
            // open configuration form, process changes after closing.
            // Not using a configuration yet, nothing to do.
        }

        void OpenForm()
        {
            if (main?.IsDisposed ?? true) main = new Main(this);
            main?.Show();
        }

        void CloseForm()
        {
            try { main?.Close(); }
            catch (Exception x) { LogWriteLine($"ERROR: App: Close; message = \"{x.Message}\""); }
            finally { main = null; }
        }
        #endregion

        #region UNCNC software events
        // put actions and here don't forget to activate the action in the API _event() versions.

        void app_init()
        {
            // nothing to do.
        }

        // object app_Informplugin(object value) {}

        // void app_Informplugins(object value) {}

        // void app_Buttonpress(int buttonnumber, bool onscreen) {}

        // void app_Toolpathclick(double X, double Y, bool Istopview) {}

        // void app_Textfieldclick(int labelnumber, bool Ismainscreen) {}

        // void app_Textfieldtexttyped(int labelnumber, bool Ismainscreen, string text) {}

        // void app_Cyclethreadstart() {}

        // void app_Cyclethreadstop() {}
        #endregion UNCNC software events

        #region Loop related events
        bool app_LoopFirstCycle()
        {
            // initialise loop
            return true;
        }

        void app_LoopCycle()
        {
            // do one loop action
        }

        void app_LoopException(Exception x)
        {
            // loop exception handling
            LogWriteLine($"Debug: plugin: LoopException(x = \"{x.Message}\")");
        }
        #endregion

        #region User events
        public void GetStatistics()
        {
            try
            {
                var result = new Statistics();
                ReturnVal r = (ReturnVal)UC100.GetStatistics(ref result);

                LogWriteLine($"GetStatistics(ref result) = {r}; DistX = {result.DistX:0.000}; DistY = {result.DistY:0.000}; DistZ = {result.DistZ:0.000}");
            }
            catch (Exception x)
            {
                LogWriteLine($"ERROR: Plugin: GetStatistics; message = \"{x.Message}\"");
            }
        }

        public void GetPosition()
        {
            try
            {
                double XposU = 0.0;
                double YposU = 0.0;
                double ZposU = 0.0;
                double AposU = 0.0;
                double BposU = 0.0;
                double CposU = 0.0;

                int XposS = 0;
                int YposS = 0;
                int ZposS = 0;
                int AposS = 0;
                int BposS = 0;
                int CposS = 0;

                var rU = (ReturnVal)GetAxisPosition(ref XposU, ref YposU, ref ZposU, ref AposU, ref BposU, ref CposU);
                var rS = (ReturnVal)GetAxisIndex(ref XposS, ref YposS, ref ZposS, ref AposS, ref BposS, ref CposS);

                LogWriteLine($"GetAxisPosition(...) = {rU}");
                LogWriteLine($"GetAxisIndex(...) = {rS}");

                LogWriteLine($"   X = {XposU:0.000} units; X = {XposS} steps");
                LogWriteLine($"   Y = {YposU:0.000} units; Y = {YposS} steps");
                LogWriteLine($"   Z = {ZposU:0.000} units; Z = {ZposS} steps");

            }
            catch (Exception x)
            {
                LogWriteLine($"ERROR: Plugin: GetPosition; message = \"{x.Message}\"");
            }
        }

        public bool WriteMDI(string value)
        {
            bool result;
            try
            {
                result = !IsMoving && !string.IsNullOrEmpty(value);
                if (result)
                {
                    if (LogVerbose) LogWriteLine($"Plugin: WriteMDI(\"{value}\")");
                    UC.Code(value.Trim());
                }
            }
            catch (Exception x)
            {
                result = false;
                LogWriteLine($"ERROR: Plugin: WriteMDI(\"{value}\"); message = \"{x.Message}\"");
            }
            return result;
        }
        #endregion User events
    }
}
